#include <iostream>
#include <cstdlib>
#include <cstdio>
#include <sys/stat.h>
#include <sys/types.h>
#include <cmath>
#include <time.h>
#include <ctype.h>
#include <new>
#include <cstring>
using namespace std;

struct data_holder
{
	unsigned int genome_locations[5];
	char at;
	data_holder *next;
	data_holder()
	{
		at=0;
		next=NULL;
	}
};

struct node
{

	data_holder *next;
	node()
	{
		next=NULL;
	}
};


int usage();
unsigned int infilesize (char *);


int main(int argc, char **argv)
{
	srand(time(0));
	int rand_num;
	//Parsing for the use of nonstandard bases and soft masking.
	bool use_nonstd=false;
	unsigned int n=2,genome_size=0;
	bool nonstandard=false;
	
	int is_header, data, slot_num, start_spot, temp_num=0, num_of_fastas=0, file_part=0;
	char header_line[100], basename[200],tempname[200];
	unsigned char base;
	FILE *instream;
	FILE *header_file, *bfa_file, *index, *index_val;
	
	//Test open fasta files.
	if (argc<5) 
	{
		fprintf(stdout, "%d\n",argc);
		return usage();
	}
	if (strcmp(argv[1], "{") !=0)
	{
		fprintf(stderr, "******************************\n");
		fprintf(stderr, "*Error in fasta input format.*\n");
		fprintf(stderr, "******************************\n");
		return usage();
	}
	while (n<argc && strcmp(argv[n],"}" )!=0)
	{
		instream = fopen(argv[n], "r");
		if (instream == (FILE *)0)
		{
			fprintf(stderr, "*********************************************\n");
			fprintf(stderr, "*Error opening fasta file %s.\n", argv[n]);
			fprintf(stderr, "*********************************************\n");
			return usage();
		}
		genome_size=genome_size+infilesize(argv[n]);
		fclose(instream);
		n++;
		num_of_fastas++;
	}
	if (n==argc)
	{
		fprintf(stderr, "******************************\n");
		fprintf(stderr, "*Error in fasta input format.*\n");
		fprintf(stderr, "******************************\n");
		return usage();
	}
	n++;	
	strcpy(basename,argv[n]);
	strcpy(tempname,basename);
	strcat(tempname,".hdr");
	header_file=fopen(tempname, "w");
	if (header_file == (FILE *)0)
	{
		fprintf(stderr, "*********************************************\n");
		fprintf(stderr, "*Error opening fasta file %s.\n", tempname);
		fprintf(stderr, "*********************************************\n");
		return usage();
	}
	strcpy(basename,argv[n]);
	strcpy(tempname,basename);
	strcat(tempname,".bfa");
	bfa_file=fopen(tempname, "wb");
	if (bfa_file == (FILE *)0)
	{
		fprintf(stderr, "*********************************************\n");
		fprintf(stderr, "*Error opening bfa %s.\n", tempname);
		fprintf(stderr, "*********************************************\n");
		return usage();
	}
	strcpy(basename,argv[n]);
	strcpy(tempname,basename);
	strcat(tempname,".val");
	index_val=fopen64(tempname, "wb");
	if (index_val == (FILE *)0)
	{
		fprintf(stderr, "*********************************************\n");
		fprintf(stderr, "*Error opening index file %s.\n", tempname);
		fprintf(stderr, "*********************************************\n");
		return usage();
	}
	n++;
	while (n<argc)
	{
		if (strcmp(argv[n],"-N")==0)
			use_nonstd=true;
		else
		{
			fprintf(stderr, "**********************\n");
			fprintf(stderr, "*Unrecognized option.*\n");
			fprintf(stderr, "**********************\n");
			return usage();
		}
		n++;
	}
	// Determine height of TRIE and store it in "HEIGHT".

	genome_size= static_cast <unsigned int> ( log(genome_size)/log(4));
	const int HEIGHT=genome_size;
	if (HEIGHT>15)
	{
		fprintf(stdout, "Cannot align genomes larger than 4 billion base pairs\n");
		return 0;
	}
	index=fopen64(basename, "wb");
	if (index == (FILE *)0)
	{
		fprintf(stderr, "*********************************************\n");
		fprintf(stderr, "*Error opening index file %s.\n", basename);
		fprintf(stderr, "*********************************************\n");
		return usage();
	}
	
	unsigned int leaf[HEIGHT], chr_loc=1, genome_loc=0, TRIE_loc=0, zero=0, greaterthan=0, file_val=0;
	int leaf_pos,count, zeros[4];
	unsigned int *loc_from_temp=new unsigned int;
	char *ws_from_temp=new char;
	bool empty=true;
	
	//Get the horizontal size of the TRIE
	int size=1, fastas=2;
		for (int a=0; a<(HEIGHT-1); a++)
			size=4*size;
		
	unsigned const int SIZE=size;
	const int SLOTS=SIZE/500000+1;
	
	//Array of pointers to store locations of data_holders.
	node *pointer[SLOTS];
	for (int a=0; a<SLOTS; a++)
		pointer[a] = new node[500000];
	//Array of 4^n powersto replace power function
	unsigned int power[HEIGHT][4];
	zero=1;
	for (temp_num=0; temp_num<HEIGHT; temp_num++)
	{
		power[temp_num][0]=0;
		power[temp_num][1]=zero;
		power[temp_num][2]=2*zero;
		power[temp_num][3]=3*zero;
		zero=4*zero;
	}
	temp_num=zero=0;
	//Temporary data_holder to be able to read info from malloc created data holders
	data_holder *temp=new data_holder, *temp2= new data_holder;
	
	//Record HEIGHT and SIZE and REF_LEN in outfile to be used in alignment.
	fprintf(header_file, "%u\n%u\n%d\n",HEIGHT, SIZE,num_of_fastas);
		
	//Build the index one quater at a time.
	for (int first_base=0; first_base<4; first_base++)
	{
		zeros[first_base]=file_val;
		fastas=2;
		genome_loc=0;
		chr_loc=1;
		greaterthan=0;
		while (strcmp(argv[fastas],"}" )!=0)
		{
			instream = fopen(argv[fastas], "r");
			getcwd(tempname, 199);
			if (first_base==0)
				fprintf(header_file,"%s/%s\n",tempname,argv[fastas]);
			//Do not use header lines from fasta files as sequence but record in the output file.
			do 
			{
				is_header=getc(instream);
				if(is_header != EOF)
					if ((is_header==';') || (is_header=='>'))
		
						fgets(header_line, 100, instream);
					if (is_header=='>' && first_base==0)
						fprintf(header_file, "%s",header_line);
		
			}
			while ((is_header == ';') || (is_header=='>') || (is_header==1));
				
			//The first base in the sequence is already stored in is_header.  Get its value.
		
			is_header = toupper(is_header);
			switch (is_header) 
			{
				case 'A': 
				leaf[0] = 0;
				break;
				case 'C':
				leaf[0] = 1;
				break;
				case 'G':
				leaf[0] = 2;
				break;
				case 'T':
				leaf[0] = 3;
				break;
				case 'N':
				if (use_nonstd==true)
				leaf[0]=rand()%4;
				else
				nonstandard=true;
				break;
				default:
			nonstandard=true;
			}
			
			//Get the sequence of the first leaf.
			n=1;
			while (n < HEIGHT)
			{
				if (!feof(instream))
					data = toupper(fgetc(instream));
		
				switch (data) 
				{
					case 'A':
					leaf[n] = 0;
					break;
					case 'C':
					leaf[n] = 1;
					break;
					case 'G':
					leaf[n] = 2;
					break;
					case 'T':
					leaf[n] = 3;
					break;
					case 'N':
					if (use_nonstd==true)
					leaf[n]=rand()%4;
					else
					nonstandard=true;
					break;
					default:
					nonstandard=true;
				}
				
				n++;
			}
			
			
			//Go through the rest of the file. 
			while (!feof(instream)) 
			{
				//Find the location of the current leaf in the TRIE.
				leaf_pos=HEIGHT;
				TRIE_loc=0;
				if (nonstandard==false && leaf[0]==first_base)
				{
					for (leaf_pos=1; leaf_pos<HEIGHT; leaf_pos++)
						TRIE_loc= (TRIE_loc+power[leaf_pos-1][leaf[HEIGHT-leaf_pos]]);

					slot_num=TRIE_loc/500000;
					start_spot=TRIE_loc%500000;
			
					//Record the location of the leaf 
					if ((pointer[slot_num]+start_spot)->next==NULL)
						((pointer[slot_num]+start_spot)->next)=new data_holder;
					temp=(pointer[slot_num]+start_spot)->next;
					while (temp->at == 5)
					{
						if (temp->next==NULL)
							temp->next=new data_holder;
						temp=temp->next;
					}
					temp->genome_locations[temp->at]=chr_loc+genome_loc;
					temp->at++;
				}
				nonstandard=false;
				
				//Get the sequence of the next leaf: 1.) Read the next HEIGHT bases into leaf
				for (count=0; (count < HEIGHT && !feof(instream)); count++)
				{
					data=fgetc(instream);
					if (data=='\n') 		//Get DNA base character
						data=fgetc(instream);	//instead of the end of line
									//character.
					data = toupper(data);
			
					switch (data) {
					case 'A':
					leaf[count] = 0;
					break;
					case 'C':
					leaf[count] = 1;
					break;
					case 'G':
					leaf[count] = 2;
					break;
					case 'T':
					leaf[count] = 3;
					break;
					case 'N':
					if (use_nonstd==true)
					leaf[count]=rand()%4;
					else
					nonstandard=true;
					break;
					default:
					nonstandard=true;}
				}
				
				//Increment genome_loc since the position is different and
				//decrement nonstandard.
				chr_loc=chr_loc+HEIGHT;
				if ((chr_loc+genome_loc)/100000000>greaterthan)
				{
					greaterthan++;
					if (first_base==0)
						fprintf(stdout, "Index A: %d00 million bases processed.\n",greaterthan);
						
					else if (first_base==1)
						fprintf(stdout, "Index C: %d00 million bases processed.\n",greaterthan);
					else if (first_base==2)
						fprintf(stdout, "Index G: %d00 million bases processed.\n",greaterthan);
					else
						fprintf(stdout, "Index T: %d00 million bases processed.\n",greaterthan);
				}
			}
			chr_loc=chr_loc+count-2;
			if (first_base==0)
				fprintf(header_file, "%u\n",chr_loc);
			genome_loc=genome_loc+chr_loc;
			chr_loc=1;
			fastas++;
			fclose(instream);
		}
		for (int y=0; y<SIZE; y++)
		{
			if ((pointer[y/500000]+(y%500000))->next != NULL)
			{
				temp=(pointer[y/500000]+(y%500000))->next;
				(pointer[y/500000]+(y%500000))->next=NULL;
				while (temp->next !=NULL)
				{
					for (int count=0; count<5; count++)
						fwrite (&temp->genome_locations[count], 4, 1, index);
					file_val+=5;
					temp2=temp;
					temp=temp->next;
					delete temp2;
				}
				for (int count=0; count<temp->at; count++)
					fwrite (&temp->genome_locations[count], 4, 1, index);
				file_val+=temp->at;
				delete temp;
			}
			fwrite(&file_val,4,1,index_val);
			if (y%10000000==0)
			{
				if 	(first_base==0)
					fprintf(stdout, "Index A: %u/%u combinations printed\n",y, SIZE);
				else if (first_base==1)
					fprintf(stdout, "Index C: %u/%u combinations printed\n",y, SIZE);
				else if (first_base==2)
					fprintf(stdout, "Index G: %u/%u combinations printed\n",y, SIZE);
				else 
					fprintf(stdout, "Index T: %u/%u combinations printed\n",y, SIZE);
			}
		}
	}
	for (int a=0; a<4; a++)
		fprintf(header_file, "%u\n",zeros[a]);
	fprintf(header_file, "%u\n", file_val);
	fclose(header_file);
	fclose(index);
	fprintf(stdout, "%u bases processed. Building compressed fasta.\n", genome_loc);
	fastas=2;
	instream = fopen(argv[fastas], "r");
	while (data=fgetc(instream) != '\n');//Read through header line.
	while (strcmp(argv[fastas],"}" )!=0)
	{
		base=0;
		for (int count =3; (count >=0 && strcmp(argv[fastas], "}")!=0) ; count--)
		{
			data=toupper(fgetc(instream));
			if (data=='\n')
				data=toupper(fgetc(instream));
			if (data==EOF)
			{
				fclose(instream);
				fastas++;
				if (strcmp(argv[fastas], "}")!=0)
				{
					instream = fopen(argv[fastas], "r");
					while (data=fgetc(instream)!='\n');
					data=toupper(fgetc(instream));
				}
			}
			switch (data)
			{
				case 'A':
					break;
				case 'C':
					base=base+power[count][1];
					break;
				case 'G':
					base=base+power[count][2];
					break;
				case 'T':
					base=base+power[count][3];
					break;
				default:
					base=base+(rand()%4)*power[count][1];
			}	
		}
		fwrite(&base,1,1,bfa_file);
		
	}
	fprintf(stdout, "Compressed fasta built.\n");
	fclose(bfa_file);
	return 0;
}

//Give usage options
int usage()
{
	fprintf(stderr, "\nUsage:\t buildindex { infile_1.fa infile_2.fa ... infile_n.fa } index.sqn [options]\n\n");
	fprintf(stderr, "\t\t-N: Treat nonstandard nuclotides as random bases.[off]\n\n");
	return 1;
}
// determine and return the file size.
unsigned int infilesize (char *filename)
{
	struct stat buf;
	int val=0;
	if (stat(filename, &buf) !=0)
		return 0;
	else val=buf.st_size;
	return (val);
}


